"""References to be imported and injected throughout the package."""
from nigsp.due import Doi

PRETI_2019 = Doi("10.1038/s41467-019-12765-7")

GRIFFA_2022 = Doi("10.1016/j.neuroimage.2022.118970")

NIGSP = Doi("10.5281/zenodo.6373436")
